﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;

namespace Infragistics.Samples.Data
{
    // from https://stackoverflow.com/questions/670577/observablecollection-doesnt-support-addrange-method-so-i-get-notified-for-each
    // allows clearing and replacing the contents of an ObservableCollection with only one event
    public class RangeCollection<T> : ObservableCollection<T>
    {
        #region Members

        public event EventHandler<ItemAddedEventArgs<T>> ItemAdded;
        public event EventHandler<ItemInsertedEventArgs<T>> ItemInserted;
        public event EventHandler<ItemRemovedEventArgs<T>> ItemRemoved;
        public event EventHandler<ItemReplacedEventArgs<T>> ItemReplaced;
        public event EventHandler<ItemsAddedEventArgs<T>> ItemsAdded;
        public event EventHandler<ItemsRemovedEventArgs<T>> ItemsRemoved;
        public event EventHandler<ItemsReplacedEventArgs<T>> ItemsReplaced;
        public event EventHandler<ItemsClearedEventArgs<T>> ItemsCleared;
        public event EventHandler<CollectionReplacedEventArgs<T>> CollectionReplaced;

        public string Id { get; set; }

        #endregion

        #region Helper Methods

        private bool HasNull(params T[] items)
        {
            foreach (var item in items)
                if (item == null)
                    return true;
            return false;
        }

        private bool HasNull(IEnumerable<T> items)
        {
            return items == null;
        }

        private bool HasNull(IEnumerable<IEnumerable<T>> items)
        {
            return items == null;
        }

        private void RaiseChanged(NotifyCollectionChangedAction action)
        {
            OnPropertyChanged(new PropertyChangedEventArgs("Count"));
            OnPropertyChanged(new PropertyChangedEventArgs("Item[]"));
            try
            {
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            }
            catch (Exception)
            {
                // some type of initialization problem
            }

        }

        #endregion

        #region Bulk Methods

        public void AddRange(IEnumerable<T> newItems)
        {
            if (HasNull(newItems)) return;
            foreach (var i in newItems) Items.Add(i);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnItemsAdded(new ItemsAddedEventArgs<T>(newItems));
        }

        public void AddRange(params IEnumerable<T>[] newItems)
        {
            if (HasNull(newItems)) return;
            foreach (var items in newItems)
            foreach (var item in items)
                this.Items.Add(item);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
        }

        public void Remove(IEnumerable<T> oldItems)
        {
            if (HasNull(oldItems)) return;
            foreach (var i in oldItems) Items.Remove(i);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnItemsRemoved(new ItemsRemovedEventArgs<T>(oldItems));
        }

        public void RemoveAll(IEnumerable<T> oldItems)
        {
            if (HasNull(oldItems)) return;
            var set = new HashSet<T>(oldItems);
            var i = 0;
            while (i < Count)
                if (set.Contains(this[i])) RemoveAt(i);
                else i++;
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnItemsRemoved(new ItemsRemovedEventArgs<T>(oldItems));
        }

        public void ReplaceAll(T old, T @new)
        {
            if (HasNull(old, @new)) return;
            Replace(old, @new, false);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnItemReplaced(new ItemReplacedEventArgs<T>(old, @new));
        }

        public void ReplaceCollection(IEnumerable<T> newItems, bool supressEvent = false)
        {
            if (HasNull(newItems)) return;
            IEnumerable<T> oldItems = new List<T>(Items);
            Items.Clear();
            foreach (var item in newItems) Items.Add(item);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnReplaced(new CollectionReplacedEventArgs<T>(oldItems, newItems));
        }

        private void Replace(T old, T @new, bool breakFirst)
        {
            var cloned = new List<T>(Items);
            var i = 0;
            foreach (var item in cloned)
            {
                if (item.Equals(old))
                {
                    Items.Remove(item);
                    Items.Insert(i, @new);
                    if (breakFirst) break;
                }

                i++;
            }
        }

        public void Replace(T old, T @new)
        {
            if (HasNull(old, @new)) return;
            Replace(old, @new, true);
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnItemReplaced(new ItemReplacedEventArgs<T>(old, @new));
        }

        #endregion

        #region New Methods

        public new void Remove(T item)
        {
            if (HasNull(item)) return;
            base.Remove(item);
            OnItemRemoved(new ItemRemovedEventArgs<T>(item));
        }

        public new void RemoveAt(int i)
        {
            var oldItem = Items[i]; 
            base.RemoveAt(i);
            OnItemRemoved(new ItemRemovedEventArgs<T>(oldItem));
        }

        public new void Clear()
        {
            IEnumerable<T> oldItems = new List<T>(Items);
            Items.Clear();
            RaiseChanged(NotifyCollectionChangedAction.Reset);
            OnCleared(new ItemsClearedEventArgs<T>(oldItems));
        }

        public new void Add(T item)
        {
            if (HasNull(item)) return;
            base.Add(item);
            OnItemAdded(new ItemAddedEventArgs<T>(item));
        }

        public new void Insert(int i, T item)
        {
            if (HasNull(item)) return;
            base.Insert(i, item);
            OnItemInserted(new ItemInsertedEventArgs<T>(item, i));
        }

        public new IEnumerable<string> ToString()
        {
            foreach (var item in this) yield return item.ToString();
        }

        #endregion

        #region Event Methods

        private void OnItemAdded(ItemAddedEventArgs<T> i)
        {
            ItemAdded?.Invoke(this, new ItemAddedEventArgs<T>(i.NewItem));
        }

        private void OnItemInserted(ItemInsertedEventArgs<T> i)
        {
            ItemInserted?.Invoke(this, new ItemInsertedEventArgs<T>(i.NewItem, i.Index));
        }

        private void OnItemRemoved(ItemRemovedEventArgs<T> i)
        {
            ItemRemoved?.Invoke(this, new ItemRemovedEventArgs<T>(i.OldItem));
        }

        private void OnItemReplaced(ItemReplacedEventArgs<T> i)
        {
            ItemReplaced?.Invoke(this, new ItemReplacedEventArgs<T>(i.OldItem, i.NewItem));
        }

        private void OnItemsAdded(ItemsAddedEventArgs<T> i)
        {
            ItemsAdded?.Invoke(this, new ItemsAddedEventArgs<T>(i.NewItems));
        }

        private void OnItemsRemoved(ItemsRemovedEventArgs<T> i)
        {
            ItemsRemoved?.Invoke(this, new ItemsRemovedEventArgs<T>(i.OldItems));
        }

        private void OnItemsReplaced(ItemsReplacedEventArgs<T> i)
        {
            ItemsReplaced?.Invoke(this, new ItemsReplacedEventArgs<T>(i.OldItems, i.NewItems));
        }

        private void OnCleared(ItemsClearedEventArgs<T> i)
        {
            ItemsCleared?.Invoke(this, new ItemsClearedEventArgs<T>(i.OldItems));
        }

        private void OnReplaced(CollectionReplacedEventArgs<T> i)
        {
            CollectionReplaced?.Invoke(this, new CollectionReplacedEventArgs<T>(i.OldItems, i.NewItems));
        }

        #endregion

        #region RangeCollection

        public RangeCollection()
        {
        }

        public RangeCollection(IEnumerable<T> collection) : base(collection)
        {
        }

        public RangeCollection(List<T> list) : base(list)
        {
        }

        public RangeCollection(params T[] items)
        {
            AddRange(items);
        }

        public RangeCollection(params IEnumerable<T>[] items)
        {
            AddRange(items);
        }

        #endregion
    }

    public class CollectionReplacedEventArgs<T> : ReplacedEventArgs<T>
    {
        public CollectionReplacedEventArgs(IEnumerable<T> old, IEnumerable<T> @new) : base(old, @new)
        {
        }
    }

    public class ItemAddedEventArgs<T> : EventArgs
    {
        public T NewItem;

        public ItemAddedEventArgs(T t)
        {
            NewItem = t;
        }
    }

    public class ItemInsertedEventArgs<T> : EventArgs
    {
        public int Index;
        public T NewItem;

        public ItemInsertedEventArgs(T t, int i)
        {
            NewItem = t;
            Index = i;
        }
    }

    public class ItemRemovedEventArgs<T> : EventArgs
    {
        public T OldItem;

        public ItemRemovedEventArgs(T t)
        {
            OldItem = t;
        }
    }

    public class ItemReplacedEventArgs<T> : EventArgs
    {
        public T NewItem;
        public T OldItem;

        public ItemReplacedEventArgs(T old, T @new)
        {
            OldItem = old;
            NewItem = @new;
        }
    }

    public class ItemsAddedEventArgs<T> : EventArgs
    {
        public IEnumerable<T> NewItems;

        public ItemsAddedEventArgs(IEnumerable<T> t)
        {
            NewItems = t;
        }
    }

    public class ItemsClearedEventArgs<T> : RemovedEventArgs<T>
    {
        public ItemsClearedEventArgs(IEnumerable<T> old) : base(old)
        {
        }
    }

    public class ItemsRemovedEventArgs<T> : RemovedEventArgs<T>
    {
        public ItemsRemovedEventArgs(IEnumerable<T> old) : base(old)
        {
        }
    }

    public class ItemsReplacedEventArgs<T> : ReplacedEventArgs<T>
    {
        public ItemsReplacedEventArgs(IEnumerable<T> old, IEnumerable<T> @new) : base(old, @new)
        {
        }
    }

    public class RemovedEventArgs<T> : EventArgs
    {
        public IEnumerable<T> OldItems;

        public RemovedEventArgs(IEnumerable<T> old)
        {
            OldItems = old;
        }
    }

    public class ReplacedEventArgs<T> : EventArgs
    {
        public IEnumerable<T> NewItems;
        public IEnumerable<T> OldItems;

        public ReplacedEventArgs(IEnumerable<T> old, IEnumerable<T> @new)
        {
            OldItems = old;
            NewItems = @new;
        }
    }
}